'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _exists = require('./utils/exists');

var _exists2 = _interopRequireDefault(_exists);

var _set = require('./utils/set');

var _set2 = _interopRequireDefault(_set);

var _get = require('./utils/get');

var _get2 = _interopRequireDefault(_get);

var _joinKeyPath = require('./utils/joinKeyPath');

var _joinKeyPath2 = _interopRequireDefault(_joinKeyPath);

var _parseKeyPath = require('./utils/parseKeyPath');

var _parseKeyPath2 = _interopRequireDefault(_parseKeyPath);

var _stringifyKeyPath = require('./utils/stringifyKeyPath');

var _stringifyKeyPath2 = _interopRequireDefault(_stringifyKeyPath);

var _buildNotificationKeyPaths = require('./utils/buildNotificationKeyPaths');

var _buildNotificationKeyPaths2 = _interopRequireDefault(_buildNotificationKeyPaths);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var isRootKeyPath = function isRootKeyPath(keyPath) {
  return !(0, _exists2.default)(keyPath) || Array.isArray(keyPath) && keyPath.length === 0;
};

var normalizeSubscriptionArgs = function normalizeSubscriptionArgs(keyPath, handler) {
  var name = arguments.length <= 2 || arguments[2] === undefined ? 'subscribe' : arguments[2];

  if (typeof keyPath === 'function') {
    handler = keyPath;
    keyPath = undefined;
  }

  (0, _invariant2.default)(typeof handler === 'function', 'Subscription handler should be a function.');

  if (!(0, _exists2.default)(keyPath)) {
    keyPath = [];
  }

  (0, _invariant2.default)(Array.isArray(keyPath) || typeof keyPath === 'string', name + ' requires array or string for key path');

  return [keyPath, handler];
};

var createTrackable = function createTrackable(get, parentKeyPath, hotSet) {
  return function (childKeyPath) {
    childKeyPath = Array.isArray(childKeyPath) ? childKeyPath : (0, _parseKeyPath2.default)(childKeyPath);
    childKeyPath = (0, _stringifyKeyPath2.default)(childKeyPath);
    var keyPath = (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath);
    hotSet[keyPath] = true;
    return get(keyPath);
  };
};

var createChildStream = function createChildStream(parentStream, parentKeyPath) {
  var get = function get(childKeyPath) {
    var fullKeyPath = (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath);
    return parentStream.get(fullKeyPath);
  };

  var set = function set(childKeyPath, childValue) {
    var fullKeyPath = (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath);
    return parentStream.set(fullKeyPath, childValue);
  };

  var subscribe = function subscribe(childKeyPath, handler) {
    if (typeof childKeyPath === 'function') {
      handler = childKeyPath;
      childKeyPath = undefined;
    }
    var fullKeyPath = (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath);
    return parentStream.subscribe(fullKeyPath, handler);
  };

  var autoSubscribe = function autoSubscribe(childKeyPath, handler) {
    if (typeof childKeyPath === 'function') {
      handler = childKeyPath;
      childKeyPath = undefined;
    }
    var fullKeyPath = (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath);
    return parentStream.autoSubscribe(fullKeyPath, handler);
  };

  var childStream = {
    get: get,
    set: set,
    subscribe: subscribe,
    autoSubscribe: autoSubscribe
  };

  childStream.at = function (childKeyPath) {
    return createChildStream(parentStream, (0, _joinKeyPath2.default)(parentKeyPath, childKeyPath));
  };

  return childStream;
};

var createStream = function createStream() {
  var _computations = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var currentValue = null;
  var computedValues = {};
  var computations = {};

  Object.keys(_computations).forEach(function (computationKey) {
    computations[(0, _stringifyKeyPath2.default)(computationKey)] = _computations[computationKey];
  });

  var get = void 0;
  var subscribe = void 0;
  var autoSubscribe = void 0;

  var pathToSubscribers = {};

  var hasComputation = function hasComputation(computationPath) {
    return computationPath in computations && typeof computations[computationPath] === 'function';
  };

  var hasHotComputation = function hasHotComputation(computationPath) {
    return(
      // If we have a cached value, then it's hot.
      computationPath in computedValues
    );
  };

  // Get value or cached computed value at path.
  var getValueAt = function getValueAt(keyPath) {
    var nullOrKeyPath = keyPath === '' ? null : keyPath;
    var value = (0, _get2.default)(currentValue, nullOrKeyPath);
    if (typeof value !== 'undefined') {
      return value;
    }
    return computedValues[keyPath];
  };

  var notifySubscribers = function notifySubscribers(keyPath) {
    keyPath = (0, _stringifyKeyPath2.default)(keyPath);

    var notificationPaths = (0, _buildNotificationKeyPaths2.default)(keyPath);

    var prefixedNotificationPaths = notificationPaths.map(function (notificationPath) {
      return 'self::' + notificationPath;
    }).concat('children::' + notificationPaths[notificationPaths.length - 1]);

    prefixedNotificationPaths.forEach(function (notificationPath) {
      var keyPathSubscribers = pathToSubscribers[notificationPath];
      if (keyPathSubscribers) {
        keyPathSubscribers.forEach(function (subscriber) {
          subscriber(getValueAt(keyPath));
        });
      }
    });
  };

  var heatUpComputation = function heatUpComputation(computationPath) {
    if (hasHotComputation(computationPath)) {
      return;
    }
    var computation = computations[computationPath];

    var isComputed = false;

    var unsubscribe = autoSubscribe(function (computationGet) {
      // First time, we're just getting the computed value and letting
      // autoSubscribe warm up.
      if (!isComputed) {
        computedValues[computationPath] = computation(computationGet);
        isComputed = true;
        return;
      }
      var prefixedComputationPath = 'self::' + computationPath;
      // After that, notify subscribers if there are any. Otherwise, just drop
      // the cache.
      if (pathToSubscribers[prefixedComputationPath]) {
        computedValues[computationPath] = computation(computationGet);
        notifySubscribers(computationPath);
      } else {
        unsubscribe();
        delete computedValues[computationPath];
      }
    });
  };

  subscribe = function subscribe(_keyPath, _handler) {
    var _normalizeSubscriptio = normalizeSubscriptionArgs(_keyPath, _handler);

    var _normalizeSubscriptio2 = _slicedToArray(_normalizeSubscriptio, 2);

    var keyPath = _normalizeSubscriptio2[0];
    var handler = _normalizeSubscriptio2[1];


    keyPath = Array.isArray(keyPath) ? keyPath : (0, _parseKeyPath2.default)(keyPath);

    keyPath = (0, _stringifyKeyPath2.default)(keyPath);

    var notificationPaths = (0, _buildNotificationKeyPaths2.default)(keyPath);

    var prefixedNotificationPaths = notificationPaths.slice(0, notificationPaths.length - 1).map(function (notificationPath) {
      return 'children::' + notificationPath;
    }).concat('self::' + notificationPaths[notificationPaths.length - 1]);

    var isSubscribed = true;

    //subscribers.push(handler);
    //subscriberPaths.push(keyPath);
    prefixedNotificationPaths.forEach(function (notificationPath) {
      pathToSubscribers[notificationPath] = pathToSubscribers[notificationPath] || [];
      pathToSubscribers[notificationPath].push(handler);
    });

    if (hasComputation(keyPath)) {
      heatUpComputation(keyPath);
    }

    return function () {
      // Don't try to unsubscribe twice.
      if (!isSubscribed) {
        return false;
      }

      isSubscribed = false;

      //const index = subscribers.indexOf(handler);
      //subscribers.splice(index, 1);
      //subscriberPaths.splice(index, 1);
      prefixedNotificationPaths.forEach(function (notificationPath) {
        var pathIndex = pathToSubscribers[notificationPath].indexOf(handler);
        pathToSubscribers[notificationPath].splice(pathIndex, 1);
        if (pathToSubscribers[notificationPath].length === 0) {
          delete pathToSubscribers[notificationPath];
        }
      });

      return undefined;
    };
  };

  get = function get(keyPath) {
    var value = (0, _get2.default)(currentValue, keyPath);
    if (typeof value === 'undefined') {
      var computationPath = (0, _stringifyKeyPath2.default)(keyPath);
      if (hasComputation(computationPath)) {
        heatUpComputation(computationPath);
        return computedValues[computationPath];
      }
    }
    return value;
  };

  var set = function set(keyPath, newValue) {

    (0, _invariant2.default)(Array.isArray(keyPath) || typeof keyPath === 'string', 'set requires array or string for key path');

    if (isRootKeyPath(keyPath)) {
      currentValue = newValue;
    } else {
      (0, _set2.default)(currentValue, keyPath, newValue);
    }

    notifySubscribers(keyPath);
  };

  autoSubscribe = function autoSubscribe(_parentKeyPath, _doSideEffect) {
    var _normalizeSubscriptio3 = normalizeSubscriptionArgs(_parentKeyPath, _doSideEffect, 'autoSubscribe');

    var _normalizeSubscriptio4 = _slicedToArray(_normalizeSubscriptio3, 2);

    var parentKeyPath = _normalizeSubscriptio4[0];
    var doSideEffect = _normalizeSubscriptio4[1];


    var hotKeyPathSubscriptions = {};

    var onChange = function onChange() {
      var newHotKeyPathSet = {};
      var getAndTrack = createTrackable(get, parentKeyPath, newHotKeyPathSet);
      doSideEffect(getAndTrack);
      // Subscribe to new hot keyPaths.
      Object.keys(newHotKeyPathSet).forEach(function (keyPath) {
        if (!hotKeyPathSubscriptions[keyPath]) {
          hotKeyPathSubscriptions[keyPath] = subscribe(keyPath, onChange);
        }
      });
      // Unsubscribe from now cold keyPaths.
      Object.keys(hotKeyPathSubscriptions).forEach(function (keyPath) {
        if (!newHotKeyPathSet[keyPath]) {
          hotKeyPathSubscriptions[keyPath]();
        }
      });
    };

    onChange();

    // unsubscribe
    return function () {
      Object.keys(hotKeyPathSubscriptions).forEach(function (keyPath) {
        hotKeyPathSubscriptions[keyPath]();
      });
    };
  };

  var stream = {
    get: get,
    set: set,
    subscribe: subscribe,
    autoSubscribe: autoSubscribe
  };

  stream.at = function (keyPath) {
    return createChildStream(stream, keyPath);
  };

  return stream;
};

exports.default = createStream;